package flm.b4a.libgdx.spfx;

import java.util.ArrayList;
import java.util.HashMap;

import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.libgdx.graphics.lgFrameBuffer;
import anywheresoftware.b4a.libgdx.graphics.lgGL;
import anywheresoftware.b4a.libgdx.graphics.lgShaderProgram;
import anywheresoftware.b4a.libgdx.graphics.lgSpriteBatch;
import anywheresoftware.b4a.libgdx.graphics.lgTexture;
import anywheresoftware.b4a.objects.collections.Map;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Mesh;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.VertexAttribute;
import com.badlogic.gdx.graphics.VertexAttributes;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.glutils.GLFrameBuffer;
import com.badlogic.gdx.graphics.glutils.ShaderProgram;
import com.badlogic.gdx.math.Matrix3;
import com.badlogic.gdx.math.Matrix4;

@BA.Version(value = 1.0f)
@BA.Author(value = "F. Leneuf-Magaud (Informatix)")

@BA.ShortName("lgSpecialFX")
@BA.Events(values = { "Draw" })
public class SpecialFX
{
	private ArrayList<Effet> ChaineEffets = new ArrayList<Effet>();

	protected class Effet
	{
		public ShaderProgram Shader;
		public Map mapParams;
		public Texture[] texInputs;
		public GLFrameBuffer[] fbInputs;
		public GLFrameBuffer fbOutput;
		public String evtDraw;

		public Effet(lgShaderProgram ShaderProg, Map Params, Object[] Inputs, int NbTexCoords, lgFrameBuffer Output, String EvtPrefix)
		{
			if (ShaderProg == null || !ShaderProg.IsInitialized() || !ShaderProg.IsCompiled())
				throw new RuntimeException("Shader program not initialized or not compiled");
			Shader = ShaderProg.getInternalObject();
			mapParams = Params;
			texInputs = new Texture[Inputs.length];
			fbInputs = new GLFrameBuffer[Inputs.length];
			int NbInputs = 0; 
			for (int i = 0; i < Inputs.length; i++) {
				if (Inputs[i] != null) {
					NbInputs++;
					if (Inputs[i].getClass() == lgTexture.class)
					{
						texInputs[i] = ((lgTexture)Inputs[i]).getInternalObject();
					}
					else if (Inputs[i].getClass() == lgFrameBuffer.class)
					{
						fbInputs[i] = ((lgFrameBuffer)Inputs[i]).getInternalObject();
						texInputs[i] = fbInputs[i].getColorBufferTexture();
					}
				}
			}
			NbTexCoords = Math.min(Math.max(1, NbInputs), NbTexCoords); 
			if (Output != null && Output.IsInitialized()) {
				fbOutput = Output.getInternalObject();
				boolean CreerMesh = true;
				if (Quads.containsKey(fbOutput)) {
					Mesh Quad = Quads.get(fbOutput);
					if (Quad.getVertexAttributes().size() >= 2 + NbTexCoords)
						CreerMesh = false; // existe dj et a suffisamment d'attributs
				}
				if (CreerMesh)
					Quads.put(fbOutput, CreateQuad(0, 0, fbOutput.getWidth(), 0, fbOutput.getWidth(), fbOutput.getHeight(), 0, fbOutput.getHeight(), NbTexCoords));
			}
			evtDraw = EvtPrefix.toLowerCase(BA.cul) + "_draw";
		}
	}

	static HashMap<GLFrameBuffer, Mesh> Quads = new HashMap<GLFrameBuffer, Mesh>();

	private static Mesh CreateQuad(float X1, float Y1, float X2, float Y2, float X3, float Y3, float X4, float Y4, int NbTexCoords)
	{
		//Creates a rectangular mesh
		float Blanc = Color.WHITE.toFloatBits();
		float[] Verts = new float[4 * (3 + (NbTexCoords * 2))];
		int i = 0;
		Verts[0] = X1;
		Verts[++i] = Y1;
		Verts[++i] = Blanc;
		for(int t = 0; t < NbTexCoords; t++) {
			Verts[++i] = 1;
			Verts[++i] = 1;
		}

		Verts[++i] = X2;
		Verts[++i] = Y2;
		Verts[++i] = Blanc;
		for(int t = 0; t < NbTexCoords; t++) {
			Verts[++i] = 0;
			Verts[++i] = 1;
		}

		Verts[++i] = X3;
		Verts[++i] = Y3;
		Verts[++i] = Blanc;
		for(int t = 0; t < NbTexCoords; t++) {
			Verts[++i] = 0;
			Verts[++i] = 0;
		}

		Verts[++i] = X4;
		Verts[++i] = Y4;
		Verts[++i] = Blanc;
		for(int t = 0; t < NbTexCoords; t++) {
			Verts[++i] = 1;
			Verts[++i] = 0;
		}

		VertexAttribute[] VA = new VertexAttribute[NbTexCoords + 2];
		VA[0] = new VertexAttribute(VertexAttributes.Usage.Position, 2, ShaderProgram.POSITION_ATTRIBUTE);
		VA[1] = new VertexAttribute(VertexAttributes.Usage.ColorPacked, 4, ShaderProgram.COLOR_ATTRIBUTE);
		for(int t = 0; t < NbTexCoords; t++) {
			VA[t+2] = new VertexAttribute(VertexAttributes.Usage.TextureCoordinates, 2, ShaderProgram.TEXCOORD_ATTRIBUTE + t);
		}
		Mesh Quad = new Mesh(true, 4, 6, VA);
		Quad.setVertices(Verts);
		Quad.setIndices(new short[] {0, 1, 2, 2, 3, 0});
		return Quad;
	}

	public void ClearAllEffects()
	{
		ChaineEffets.clear();
	}

	public void dispose()
	{
		ChaineEffets.clear();
		Quads.clear();
	}

	/**
	 * Adds an effect to the effect chain.
	 * @param ShaderProg The compiled shader program to run.
	 * @param Params The parameters to pass to the shader program (to set the uniforms). May be Null.
	 * @param Input The texture or framebuffer object used as input. The texture will be bound to the appropriate target of the texture unit 0. If Null, a Draw event will be raised (the SpriteBatch parameter of the Apply function is mandatory in this case; SpriteBatch.Begin will be called before raising the event).
	 * @param Output The framebuffer object to render to. Can only be Null if Input is Null. In this case, the rendering is done to the display with the specified SpriteBatch (see Apply). 
	 * @param EvtPrefix The prefix of the Draw event.
	 */
	public void AddEffect(lgShaderProgram ShaderProg, Parameters Params, Object Input, lgFrameBuffer Output, String EvtPrefix)
	{
		Effet NouvelEffet = new Effet(ShaderProg, Params == null ? null : Params.getParamMap(), new Object[] { Input }, 1, Output, EvtPrefix); 
		ChaineEffets.add(NouvelEffet);
	}

	/**
	 * Adds an effect to the effect chain. This function allows multiple inputs.
	 * @param ShaderProg The compiled shader program to run.
	 * @param Params The parameters to pass to the shader program (uniforms). May be Null.
	 * @param Inputs The textures or framebuffer objects used as inputs. The first texture will be bound to the appropriate target of the texture unit 0, the second texture will be bound to the appropriate target of the texture unit 1, and so on.
	 * @param NbTexCoords The number of texture coordinates attributes. Cannot be greater than the number of Inputs. The first attribute is called "a_texCoord0", the second attribute is called "a_texCoord1", and so on.
	 * @param Output The framebuffer object to render to. Can only be Null if Input is Null. In this case, the rendering is done to the display with the specified SpriteBatch (see Apply). 
	 * @param EvtPrefix The prefix of the Draw event.
	 */
	public void AddEffectMultiTex(lgShaderProgram ShaderProg, Parameters Params, Object[] Inputs, int NbTexCoords, lgFrameBuffer Output, String EvtPrefix)
	{
		Effet NouvelEffet = new Effet(ShaderProg, Params == null ? null : Params.getParamMap(), Inputs, NbTexCoords, Output, EvtPrefix); 
		ChaineEffets.add(NouvelEffet);
	}

	Matrix4 Matrice2D = new Matrix4();

	/**
	 * Apply all effects of the effect chain, one after the other.
	 * @param SpriteBatch The SpriteBatch renderer to use in case an effect has no input. SpriteBatch.Begin is called before raising the Draw event, and SpriteBatch.End is called after.
	 * This function may change the SpriteBatch shader program, so think to reset it after the call. 
	 */
	public void Apply(BA ba, lgSpriteBatch SpriteBatch)
	{
		SpriteBatch Batch = null;
		if (SpriteBatch != null) {
			Batch = SpriteBatch.getInternalObject();
		}
		for (int e = 0; e < ChaineEffets.size(); e++)
		{
			// Rendu de la scne avec l'effet spcifi
			Effet FX = ChaineEffets.get(e);
			if (FX.fbOutput != null) {
				FX.fbOutput.begin();
			}
			if (FX.texInputs[0] == null)
			{
				// Rendu avec SpriteBatch en appelant l'vnement Draw
				Batch.setShader(FX.Shader);
				Batch.begin();
				if (FX.fbOutput != null) {
					Matrice2D.setToOrtho2D(0, 0, FX.fbOutput.getWidth(), FX.fbOutput.getHeight());
					FX.Shader.setUniformMatrix("u_projTrans", Matrice2D);
				}
				AppliquerParams(FX.Shader, FX.mapParams);
				ba.raiseEvent(this, FX.evtDraw, new Object[0]);
				Batch.end();
			}
			else if (FX.fbOutput != null)
			{
				// Rendu avec un quad
				for (int i = FX.texInputs.length - 1; i >= 0; i--) {
					FX.texInputs[i].bind(i);
				}
				FX.Shader.begin();
				Matrice2D.setToOrtho2D(0, 0, FX.fbOutput.getWidth(), FX.fbOutput.getHeight());
				FX.Shader.setUniformMatrix("u_projTrans", Matrice2D);
				FX.Shader.setUniformi("u_texture", 0);
				AppliquerParams(FX.Shader, FX.mapParams);
				Mesh Quad = Quads.get(FX.fbOutput);
				Quad.render(FX.Shader, lgGL.GL10_TRIANGLES);
				FX.Shader.end();
			}
			if (FX.fbOutput != null) {
				FX.fbOutput.end();
			}
		}
	}

	private void AppliquerParams(ShaderProgram Shader, Map Params)
	{
		if (Params == null || !Params.IsInitialized()) return;
		for (int i = 0; i < Params.getSize(); i++)
		{
			String Nom = (String) Params.GetKeyAt(i);
			Class<?> clsValeur = Params.GetValueAt(i).getClass();
			if (clsValeur == int[].class)
			{
				int[] intParam = (int[]) Params.GetValueAt(i);
				switch (intParam.length)
				{
				case 1:
					Shader.setUniformi(Nom, intParam[0]);
					break;
				case 2:
					Shader.setUniformi(Nom, intParam[0], intParam[1]);
					break;
				case 3:
					Shader.setUniformi(Nom, intParam[0], intParam[1], intParam[2]);
					break;
				case 4:
					Shader.setUniformi(Nom, intParam[0], intParam[1], intParam[2], intParam[3]);
					break;
				}
			}
			else if (clsValeur == float[].class)
			{
				float[] floatParam = (float[]) Params.GetValueAt(i);
				switch (floatParam.length)
				{
				case 1:
					Shader.setUniformf(Nom, floatParam[0]);
					break;
				case 2:
					Shader.setUniformf(Nom, floatParam[0], floatParam[1]);
					break;
				case 3:
					Shader.setUniformf(Nom, floatParam[0], floatParam[1], floatParam[2]);
					break;
				case 4:
					Shader.setUniformf(Nom, floatParam[0], floatParam[1], floatParam[2], floatParam[3]);
					break;
				}
			}
			else if (clsValeur == Matrix3.class)
			{
				Matrix3 m3Param = (Matrix3) Params.GetValueAt(i);
				Shader.setUniformMatrix(Nom, m3Param);
			}
			else if (clsValeur == Matrix4.class)
			{
				Matrix4 m4Param = (Matrix4) Params.GetValueAt(i);
				Shader.setUniformMatrix(Nom, m4Param);
			}
		}
	}
}
